package fr.lri.keller.td4serverui;

import android.os.AsyncTask;
import android.support.v7.app.AppCompatActivity;
import android.os.Bundle;
import android.view.Menu;
import android.view.MenuItem;
import android.view.View;
import android.widget.ScrollView;
import android.widget.TextView;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.util.Scanner;
import java.util.concurrent.ConcurrentLinkedQueue;

public class ServerActivity extends AppCompatActivity {

    private TextView chat;
    private ScrollView scroll;
    private String chatLog = "";

    private ServerThread serverThread = new ServerThread();
    private boolean serverRunning = true;
    private ServerSocket server;
    private Socket client;

    // Il y a plusieurs choix d'implantation possibles :
    // - une seule tâche de fond, avec plusieurs boucles imbriquées pour attendre en permanence un
    //   client, puis écouter en permanence sur le flux entrant lorsqu'un client s'est connecté
    //   (choix effectué ici)
    // - plusieurs tâches de fond, une par boucle d'attente, qui se lancent les unes les autres au
    //   fur et à mesure de l'exécution
    private class ServerThread extends AsyncTask<Void,String,Void> {
        @Override
        protected Void doInBackground(Void... voids) {
            // Ouverture de la socket serveur
            int port = 7777;
            try {
                server = new ServerSocket(port);
            } catch (Exception e) {
                publishProgress("Impossible de lancer le serveur");
                return null;
            }
            publishProgress("Serveur lancé");
            // Boucle infinie pour attendre un client, qui est arrêtée lorsque la tâche de fond
            // l'est
            while (!isCancelled()) {
                try {
                    client = server.accept();
                    publishProgress("Connexion d'un client");
                    PrintWriter out = new PrintWriter(client.getOutputStream(), true);
                    BufferedReader in = new BufferedReader(new InputStreamReader(client.getInputStream()));
                    // On s'attend d'abord à avoir un login
                    String login = null;
                    while (login == null) {
                        String message = in.readLine();
                        login = getLogin(message);
                        if (login == null) { out.println("Login incorrect"); }
                    }
                    String welcome = "Welcome " + login;
                    publishProgress(welcome);
                    out.println(welcome);
                    // On traite ensuite en continu les messages du client, jusqu'à recevoir un
                    // logout ou un message incorrect
                    while (true) {
                        String message = in.readLine();
                        String sol = getSendOrLogout(message);
                        if (sol == null) {
                            String bye = "Bye bye " + login;
                            publishProgress(bye);
                            out.println(bye);
                            client.close();
                            break;
                        } else {
                            String send = login + ": " + sol;
                            publishProgress(send);
                            out.println(send);
                        }
                    }
                } catch (Exception e) { }
            }
            return null;
        }

        @Override
        protected void onProgressUpdate(String... messages) {
            displayMessage(messages[0]);
        }
    }

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_server);
        chat = (TextView) findViewById(R.id.chat);
        scroll = (ScrollView) findViewById(R.id.scroll);
    }

    @Override
    protected void onStart() {
        super.onStart();
        serverThread.execute();
    }

    @Override
    public boolean onCreateOptionsMenu(Menu menu) {
        // Inflate the menu; this adds items to the action bar if it is present.
        getMenuInflater().inflate(R.menu.menu_server, menu);
        return true;
    }

    @Override
    public boolean onOptionsItemSelected(MenuItem item) {
        // Handle action bar item clicks here. The action bar will
        // automatically handle clicks on the Home/Up button, so long
        // as you specify a parent activity in AndroidManifest.xml.
        int id = item.getItemId();

        //noinspection SimplifiableIfStatement
        if (id == R.id.action_settings) {
            return true;
        }

        return super.onOptionsItemSelected(item);
    }

    private void displayMessage(String message) {
        // On ajoute ce message a l'ensemble du chat
        chatLog += message + "\n";
        // On affiche le chat
        chat.setText(chatLog);
        // On scrolle en bas pour visualiser le dernier message
        scroll.fullScroll(View.FOCUS_DOWN);

    }

    private String getLogin(String message) {
        Scanner scan = new Scanner(message);
        scan.useDelimiter(" ");
        if (scan.hasNext() && scan.next().equals("LOGIN") && scan.hasNext()) {
            return scan.next();
        } else {
            return null;
        }
    }

    private String getSendOrLogout(String message) {
        Scanner scan = new Scanner(message);
        scan.useDelimiter(" ");
        if (scan.hasNext() && scan.next().equals("SEND")) {
            return message.substring(5);
        } else {
            return null;
        }
    }

    // Lorsqu'on appuie sur le bouton, on arrête le serveur s'il est lancé, et on le relance s'il
    // est arrêté
    // Utiliser "serverThread.cancel(true)" ne suffit pas : si le serveur est sur une action
    // bloquante (attente d'un message du client ou attente d'un client), il reste sur cette action
    // et on ne repasse pas tout de suite dans le test "isCancelled". Il faut donc également
    // stopper les actions bloquantes. La solution choisie ici (qui n'est pas forcément la plus
    // propre) est de fermer les sockets client et serveur lorsqu'elles sont ouvertes, ce qui va
    // provoquer une exception lors des actions bloquantes et ainsi forcer le passage dans le test
    // "isCancelled".
    public void arretLancement(View view) {
        if (serverRunning) {
            serverThread.cancel(true);
            try { client.close(); } catch (Exception e) { }
            try { server.close(); } catch (Exception e) { }
            displayMessage("Serveur arrêté");
        } else {
            serverThread = new ServerThread();
            serverThread.execute();
        }
        serverRunning = !serverRunning;
    }

}
